#!/usr/bin/env python3

import argparse
import logging
import time

import dropbox

from backup_to_dropbox.clients import DropboxClient
from backup_to_dropbox.services import BackupService, GpgEncryptionService



def main():
    parser = argparse.ArgumentParser(description='Backup data using Dropbox as storage.')
    parser.add_argument('--api-key',
                        required=True,
                        help='Dropbox API Key to use for authentication')
    parser.add_argument('--backup-name',
                        required=True,
                        help='Name for the backup in Dropbox')
    parser.add_argument('--max-backups',
                        type=int,
                        help='Max number of these backups to keep in Dropbox')
    parser.add_argument('--gpg-encrypt',
                        help='Key fingerprint to use for encrypting')
    parser.add_argument('--gpg-home',
                        help='Folder to use as GPG home')
    parser.add_argument('--gpg-pubkeyring',
                        help='GPG public key keyring to use')
    parser.add_argument('paths',
                        nargs='+',
                        help='List of paths to include in the backup')

    args = parser.parse_args()
    start_time = time.perf_counter()
    logging.info('Creating Dropbox client')
    dropbox_client = DropboxClient(dropbox.Dropbox(args.api_key, timeout=None))

    encryption_service = None
    if args.gpg_encrypt is not None:
        from pretty_bad_protocol import gnupg

        gnupg_api = gnupg.GPG(homedir=args.gpg_home, keyring=args.gpg_pubkeyring)
        encryption_service = GpgEncryptionService(args.gpg_encrypt, gnupg_api)

    backup_service = BackupService(dropbox_client, args.backup_name, encryption_service)
    if args.max_backups is not None:
        logging.info('Performing cleanup of old backups')
        backup_service.cleanup_old_backups(args.max_backups - 1)
    backup_service.backup_paths(args.paths)
    end_time = time.perf_counter()
    logging.info('Backup finished. Time elapsed: %.2f', end_time - start_time)

if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s', level=logging.INFO)

    main()



import io
import tarfile
import unittest

from datetime import datetime
from unittest.mock import ANY, MagicMock, Mock, call, patch

from backup_to_dropbox.services import BackupService

class BackupServiceTest(unittest.TestCase):

    def setUp(self):
        self.dropbox_client = Mock()
        self.encryption_service = Mock()
        self.backup_name = 'TestBackupName'
        self.backup_service = BackupService(self.dropbox_client,
                                            self.backup_name,
                                            self.encryption_service)

    def test_cleanup_not_needed(self):
        self.dropbox_client.list_files.return_value = [
            '2020-05-02-0000.tar.gz',
            '2020-05-01-0000.tar.gz',
        ]

        self.backup_service.cleanup_old_backups(2)

        self.dropbox_client.list_files.assert_called_once_with('/' + self.backup_name)

    def test_cleanup_deletes_correct_files(self):
        self.dropbox_client.list_files.return_value = [
            '2020-05-02-0000.tar.gz',
            '2020-05-01-0000.tar.gz',
            '2020-05-03-0000.tar.gz',
            '2020-04-03-0000.tar.gz',
        ]

        self.backup_service.cleanup_old_backups(2)

        self.dropbox_client.list_files.assert_called_once_with('/' + self.backup_name)
        self.dropbox_client.delete_file.assert_has_calls([
            call('/{}/2020-04-03-0000.tar.gz'.format(self.backup_name)),
            call('/{}/2020-05-01-0000.tar.gz'.format(self.backup_name)),
        ])

    @patch.object(BackupService, 'now')
    @patch.object(BackupService, 'isfile')
    @patch.object(BackupService, 'isdir')
    @patch('tarfile.open')
    def test_backup_with_encryption(self, open_tarfile_mock, isdir_mock, isfile_mock, date_now):
        backup_path = '/var/log/test'
        date_now.return_value = datetime(2020, 5, 10, 17, 15, 30)

        backup_tar_file = MagicMock()
        open_tarfile_mock.return_value.__enter__.return_value = backup_tar_file
        isfile_mock.return_value = True

        encrypted_file = Mock()
        self.encryption_service.encrypt.return_value = encrypted_file

        self.backup_service.backup_paths([backup_path])

        open_tarfile_mock.assert_called_once()
        isfile_mock.assert_called_once()
        backup_tar_file.add.assert_called_once_with(backup_path)

        self.encryption_service.encrypt.assert_called_once()

        expected_filepath = '/{}/2020-05-10-1715.tar.gz.enc'.format(self.backup_name)
        self.dropbox_client.upload_file.assert_called_once_with(encrypted_file, expected_filepath)

    @patch.object(BackupService, 'now')
    @patch.object(BackupService, 'isfile')
    @patch.object(BackupService, 'isdir')
    @patch('tarfile.open')
    def test_backup_without_encryption(self, open_tarfile_mock, isdir_mock, isfile_mock, date_now):
        self.backup_service = BackupService(self.dropbox_client, self.backup_name)

        backup_path = '/var/log/test'
        date_now.return_value = datetime(2020, 5, 10, 17, 15, 30)

        backup_tar_file = MagicMock()
        open_tarfile_mock.return_value.__enter__.return_value = backup_tar_file
        isfile_mock.return_value = True

        self.backup_service.backup_paths([backup_path])

        open_tarfile_mock.assert_called_once()
        isfile_mock.assert_called_once()
        backup_tar_file.add.assert_called_once_with(backup_path)

        expected_filepath = '/{}/2020-05-10-1715.tar.gz'.format(self.backup_name)
        self.dropbox_client.upload_file.assert_called_once_with(ANY, expected_filepath)

    @patch.object(BackupService, 'now')
    @patch.object(BackupService, 'isfile')
    @patch.object(BackupService, 'isdir')
    @patch('tarfile.open')
    def test_backup_add_existing_elems(self, open_tarfile_mock, isdir_mock, isfile_mock, date_now):
        existing_file = '/var/log/test'
        existing_dir = '/var/log/test_dir'
        non_existing_path = '/var/log/missing'

        def isfile(path):
            if path == existing_file:
                return True
            return False

        def isdir(path):
            if path == existing_dir:
                return True
            return False

        date_now.return_value = datetime(2020, 5, 10, 17, 15, 30)

        backup_tar_file = MagicMock()
        open_tarfile_mock.return_value.__enter__.return_value = backup_tar_file
        isfile_mock.side_effect = isfile

        encrypted_file = Mock()
        self.encryption_service.encrypt.return_value = encrypted_file

        self.backup_service.backup_paths([existing_file, existing_dir, non_existing_path])

        open_tarfile_mock.assert_called_once()
        backup_tar_file.add.assert_has_calls([call(existing_file), call(existing_dir)],
                                             any_order=True)

        self.encryption_service.encrypt.assert_called_once()

        expected_filepath = '/{}/2020-05-10-1715.tar.gz.enc'.format(self.backup_name)
        self.dropbox_client.upload_file.assert_called_once_with(encrypted_file, expected_filepath)







import unittest

from unittest.mock import Mock, call, ANY

from backup_to_dropbox.services import GpgEncryptionService

class GpgEncryptionServiceTest(unittest.TestCase):
    def setUp(self):
        self.destination = 'TARGET_DESTINATION'
        self.gnupg_mock = Mock()
        self.encryption_service = GpgEncryptionService(self.destination, self.gnupg_mock)

    def test_successful_encryption(self):
        file_to_encrypt = Mock()
        encrypted_file_contents = b'encrypted value'

        def encrypt_file(file_in, target, output, armor):
            with open(output, 'wb') as file_out:
                file_out.write(encrypted_file_contents)
            return Mock(ok=True)

        self.gnupg_mock.encrypt.side_effect = encrypt_file

        file_out = self.encryption_service.encrypt(file_to_encrypt)
        file_out.seek(0)
        content = file_out.read()
        self.assertEqual(encrypted_file_contents, content)
        file_out.close()

        self.gnupg_mock.encrypt.assert_called_once_with(file_to_encrypt,
                                                        self.destination,
                                                        output=ANY,
                                                        armor=False)

    def test_fail_to_encrypt_raises_error(self):
        file_to_encrypt = Mock()

        self.gnupg_mock.encrypt.return_value = Mock(ok=False, status='Could not encrypt')

        self.assertRaises(Exception, self.encryption_service.encrypt, file_to_encrypt)

        self.gnupg_mock.encrypt.assert_called_once_with(file_to_encrypt,
                                                        self.destination,
                                                        output=ANY,
                                                        armor=False)




import logging
import os
import os.path
import tarfile

from datetime import datetime
from tempfile import TemporaryDirectory, NamedTemporaryFile, TemporaryFile



class BackupService:

    def __init__(self, dropbox_client, backup_name, encryption_service=None):
        self.__dropbox_client = dropbox_client
        self.__base_dir = os.path.join('/', backup_name)
        self.__encryption_service = encryption_service

    def backup_paths(self, paths):
        execution_time = BackupService.now()
        backup_file = self._generate_backup_file(paths)
        filename = '{}.tar.gz'.format(execution_time.strftime(r'%Y-%m-%d-%H%M'))

        if self.__encryption_service is not None:
            backup_file = self.__encryption_service.encrypt(backup_file)
            filename += '.enc'

        logging.info('Uploading backup: %s', self._get_dropbox_path(filename))
        self.__dropbox_client.upload_file(backup_file, self._get_dropbox_path(filename))


    def cleanup_old_backups(self, max_to_keep):
        current_backups = self.__dropbox_client.list_files(self.__base_dir)

        if len(current_backups) > max_to_keep:
            files_to_delete = len(current_backups) - max_to_keep
            logging.info('Found %d backups: removing the oldes %d',
                         len(current_backups),
                         files_to_delete)
            current_backups.sort()
            for file in current_backups[:files_to_delete]:
                logging.info('Deleting file: %s', self._get_dropbox_path(file))
                self.__dropbox_client.delete_file(self._get_dropbox_path(file))

    def _generate_backup_file(self, paths):
        raw_file = TemporaryFile()
        with tarfile.open(fileobj=raw_file, mode='w:gz') as targz_file:
            for path in filter(lambda p: BackupService.isfile(p) or BackupService.isdir(p), paths):
                logging.debug('Adding path "%s" to backup', path)
                targz_file.add(path)

        return raw_file

    def _get_dropbox_path(self, filename):
        return os.path.join(self.__base_dir, filename)

    @staticmethod
    def now():
        return datetime.now()

    @staticmethod
    def isfile(path):
        return os.path.isfile(path)

    @staticmethod
    def isdir(path):
        return os.path.isdir(path)



class GpgEncryptionService:

    def __init__(self, destination, gpg_api):
        self.__gpg = gpg_api
        self.__dest = destination
        self.__temp_dir = TemporaryDirectory()

    def encrypt(self, fileobj_input):
        logging.info('Encrypting file using GPG encryption to key: %s', self.__dest)
        fileobj_input.seek(0)
        encrypted_file = os.path.join(self.__temp_dir.name, 'encrypted-file')
        try:
            res = self.__gpg.encrypt(fileobj_input,
                                     self.__dest,
                                     output=encrypted_file,
                                     armor=False)
            if not res.ok:
                raise Exception(res.status)
        except Exception as e:
            logging.error('Failed to encrypt backup file: {}'.format(e))
            raise

        fileobj_input.close()
        return open(encrypted_file, 'rb')



import io
import unittest

from unittest.mock import Mock, call

from dropbox.exceptions import ApiError
from dropbox.file_properties import LookupError
from dropbox.files import (CommitInfo, UploadSessionCursor, WriteMode)

from backup_to_dropbox.clients import DropboxClient



class DropboxClientTest(unittest.TestCase):
    def setUp(self):
        self.api_mock = Mock()
        self.dropbox_client = DropboxClient(self.api_mock)

    def test_file_deletion(self):
        file_to_delete = '/test/file/to/delete'
        self.dropbox_client.delete_file(file_to_delete)
        self.api_mock.files_delete_v2.assert_called_once_with(file_to_delete)

    def test_listing_files_on_empty_folder(self):
        folder_path = '/test/'

        api_call_result = Mock(entries=[])

        self.api_mock.files_list_folder = Mock(return_value=api_call_result)
        files_found = self.dropbox_client.list_files(folder_path)
        self.assertEqual([], files_found)
        self.api_mock.files_list_folder.assert_called_once_with(folder_path)

    def test_listing_files_on_folder_with_files(self):
        folder_path = '/test/'

        file_1 = Mock()
        file_1.configure_mock(name='2020-01')
        file_2 = Mock()
        file_2.configure_mock(name='2020-02')

        api_call_result = Mock(entries=[file_1, file_2])
        expected_files_found = ['2020-01', '2020-02']

        self.api_mock.files_list_folder = Mock(return_value=api_call_result)
        files_found = self.dropbox_client.list_files(folder_path)
        self.assertEqual(expected_files_found, files_found)
        self.api_mock.files_list_folder.assert_called_once_with(folder_path)

    def test_listing_files_on_non_existing_folder(self):
        folder_path = '/test/'

        lookup_error = LookupError.not_found
        folder_error = Mock()
        folder_error.is_path = Mock(return_value=True)
        folder_error.get_path = Mock(return_value=lookup_error)
        not_found_error = ApiError(request_id='1',
                                   error=folder_error,
                                   user_message_text=None,
                                   user_message_locale=None)

        self.api_mock.files_list_folder = Mock(side_effect=not_found_error)
        files_found = self.dropbox_client.list_files(folder_path)
        self.assertEqual([], files_found)
        self.api_mock.files_list_folder.assert_called_once_with(folder_path)

    def test_listing_files_other_error_is_propagated(self):
        folder_path = '/test/'

        non_folder_error = Mock()
        non_folder_error.is_path = Mock(return_value=False)
        not_found_error = ApiError(request_id='1',
                                   error=non_folder_error,
                                   user_message_text=None,
                                   user_message_locale=None)

        self.api_mock.files_list_folder = Mock(side_effect=not_found_error)
        self.assertRaises(ApiError, self.dropbox_client.list_files, folder_path)
        self.api_mock.files_list_folder.assert_called_once_with(folder_path)

    def test_upload_small_file_single_request(self):
        file_path = '/test/12345'
        file_contents = 'abcdefg'
        file_mock = Mock()
        file_mock.tell = Mock(side_effect=[15, DropboxClient.SINGLE_REQ_UPLOAD_SIZE_LIMIT])
        file_mock.read = Mock(return_value=file_contents)

        self.dropbox_client.upload_file(file_mock, file_path)
        self.api_mock.files_upload.assert_called_once_with(file_contents, file_path)
        file_mock.seek.assert_called_with(0)

    def test_upload_big_file_multiple_requests(self):
        session_id = Mock(session_id='12345')
        file_path = '/test/12345'
        file_chunks = [
            'a' * DropboxClient.SINGLE_REQ_UPLOAD_SIZE_LIMIT,
            'b' * DropboxClient.SINGLE_REQ_UPLOAD_SIZE_LIMIT,
            'c'
        ]
        file_mock = Mock()
        file_mock.tell = Mock(side_effect=[15, DropboxClient.SINGLE_REQ_UPLOAD_SIZE_LIMIT * 2 + 1])
        file_mock.read = Mock(side_effect=file_chunks)

        self.api_mock.files_upload_session_start.return_value = session_id
        self.dropbox_client.upload_file(file_mock, file_path)

        self.api_mock.files_upload_session_start.assert_called_once_with(file_chunks[0])
        self.api_mock.files_upload_session_append_v2.assert_has_calls([
            call(file_chunks[1],
                 UploadSessionCursor('12345', DropboxClient.SINGLE_REQ_UPLOAD_SIZE_LIMIT))
        ])
        self.api_mock.files_upload_session_finish.assert_has_calls([
            call(file_chunks[2],
                 UploadSessionCursor('12345', DropboxClient.SINGLE_REQ_UPLOAD_SIZE_LIMIT * 2),
                 CommitInfo(path=file_path,
                            mode=WriteMode('add'),
                            autorename=False))
        ])
        file_mock.seek.assert_called_with(0)




import io
import logging

from dropbox.exceptions import ApiError
from dropbox.files import (CommitInfo, UploadSessionCursor, WriteMode)

class DropboxClient:

    SINGLE_REQ_UPLOAD_SIZE_LIMIT = 150 * 1024 * 1024 # 150MB

    def __init__(self, dbx_api_client):
        self.__dropbox_client = dbx_api_client

    def upload_file(self, file_to_upload, path):
        file_size = self._get_file_size(file_to_upload)
        file_to_upload.seek(0)

        if file_size <= DropboxClient.SINGLE_REQ_UPLOAD_SIZE_LIMIT:
            logging.debug('Using single request to upload file')
            self.__dropbox_client.files_upload(file_to_upload.read(), path)
        else:
            logging.debug('Using multi-request upload session for this file')
            session = None
            offset = 0
            chunk = file_to_upload.read(DropboxClient.SINGLE_REQ_UPLOAD_SIZE_LIMIT)
            while len(chunk) == DropboxClient.SINGLE_REQ_UPLOAD_SIZE_LIMIT:
                if session is None:
                    logging.debug('Initializing upload session')
                    session = self.__dropbox_client.files_upload_session_start(chunk)
                else:
                    logging.debug('Appending to session %s at offset %d',
                                  session.session_id,
                                  offset)
                    self.__dropbox_client.files_upload_session_append_v2(
                        chunk,
                        UploadSessionCursor(session.session_id, offset)
                    )
                offset += len(chunk)
                chunk = file_to_upload.read(DropboxClient.SINGLE_REQ_UPLOAD_SIZE_LIMIT)

            logging.debug('Finishing session %s', session.session_id)
            commit_info = CommitInfo(path=path,
                                     mode=WriteMode('add'),
                                     autorename=False)
            self.__dropbox_client.files_upload_session_finish(
                chunk,
                UploadSessionCursor(session.session_id, offset),
                commit_info
            )

    def delete_file(self, path):
        self.__dropbox_client.files_delete_v2(path)

    def list_files(self, folder_path):
        try:
            return [entry.name for \
                    entry in self.__dropbox_client.files_list_folder(folder_path).entries]
        except ApiError as e:
            folder_list_error = e.error
            if not folder_list_error.is_path():
                raise
            lookup_error = folder_list_error.get_path()
            if lookup_error.is_not_found():
                return []
            raise

    def _get_file_size(self, file_handle):
        current_tell = file_handle.tell()
        file_handle.seek(0, io.SEEK_END)
        size = file_handle.tell()
        file_handle.seek(current_tell, io.SEEK_SET)
        return size





from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="backup-to-dropbox",
    version="0.2.0",
    packages=find_packages(),
    scripts=["backup-to-dropbox.py"],

    install_requires=[
        "dropbox==10.1.1",
        "pretty-bad-protocol==3.1.1",
    ],

    author="Santiago Alessandri",
    author_email="backup-to-dropbox@salessandri.name",
    description="Tool to generate backups (tar.gz files) and upload them to Dropbox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="backup dropbox",
    url="https://github.com/salessandri/backup-to-dropbox",
    classifiers=[
        "Development Status :: 4 - Beta",

        "Topic :: System :: Archiving :: Backup",
        "Topic :: Utilities",

        "License :: OSI Approved :: MIT License"
    ]
)
